// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"

	"gitlab.bit9.local/psc/ems/hinting/internal/config"
	"gitlab.bit9.local/psc/ems/hinting/internal/logger"
	"gitlab.bit9.local/psc/platform/orgkey-go/v3"

	"github.com/gin-gonic/gin"
	"gitlab.bit9.local/psckit/pschttp"
	"gitlab.bit9.local/psckit/pscrbac"
	"go.uber.org/zap"
	gcontext "gopkg.in/h2non/gentleman.v2/context"
	"gopkg.in/h2non/gentleman.v2/plugin"
	"gopkg.in/h2non/gentleman.v2/plugins/timeout"
)

// UserContextMiddleware -- switch function to choose between local and deployed middleware usage
//
//goland:noinspection GoMixedReceiverTypes
func (mwf RouteMiddlewareFactory) UserContextMiddleware() gin.HandlerFunc {
	if mwf.cfg.IsLocalEnv() || mwf.cfg.IsDevEnv() {
		return localUserContextMiddleware()
	} else {
		return realUserContextMiddleware()
	}
}

func localUserContextMiddleware() gin.HandlerFunc {
	userContext := pschttp.UserContext{LoginSession: pschttp.UserOrgSession{}}
	return func(c *gin.Context) {
		c.Set(pschttp.UserContextHeader, userContext)
		c.Next()
	}
}

func realUserContextMiddleware() gin.HandlerFunc {
	return func(c *gin.Context) {
		userContextHeader := c.Request.Header.Get(pschttp.UserContextHeader)
		if userContextHeader != "" {
			userContext := new(pschttp.UserContext)
			if err := json.Unmarshal([]byte(userContextHeader), userContext); err != nil {
				_ = c.AbortWithError(http.StatusInternalServerError, fmt.Errorf("unable to decode user context header %w", err))
				return
			}
			c.Set(pschttp.UserContextHeader, *userContext)
		}
		c.Next()
	}
}

func localRequestContextMiddleware() gin.HandlerFunc {
	requestContext := pschttp.RequestContext{}
	return func(c *gin.Context) {
		c.Set(pschttp.RequestContextHeader, requestContext)
		c.Next()
	}
}

func realRequestContextMiddleware() gin.HandlerFunc {
	return func(c *gin.Context) {
		requestContextHeader := c.Request.Header.Get(pschttp.RequestContextHeader)
		if requestContextHeader != "" {
			requestContext := new(pschttp.RequestContext)
			if err := json.Unmarshal([]byte(requestContextHeader), requestContext); err != nil {
				_ = c.AbortWithError(http.StatusInternalServerError, fmt.Errorf("unable to decode request context %w", err))
				return
			}
			c.Set(pschttp.RequestContextHeader, *requestContext)
		}
		c.Next()
	}
}

//goland:noinspection GoMixedReceiverTypes
func (mwf RouteMiddlewareFactory) RequestContextMiddleware() gin.HandlerFunc {
	if mwf.cfg.IsLocalEnv() || mwf.cfg.IsDevEnv() {
		return localRequestContextMiddleware()
	} else {
		return realRequestContextMiddleware()
	}
}

type demandChecker interface {
	HasPermission(ctx context.Context, auth pschttp.Authentication, caller pschttp.UserOrgSession, orgKey orgkey.OrgKey, permissions ...pscrbac.Permission) (bool, error)
}

type RbacMiddleWare struct {
	demandService demandChecker
	permissions   []pscrbac.Permission
}

type localDemandChecker struct{}

func (local localDemandChecker) HasPermission(_ context.Context, _ pschttp.Authentication, _ pschttp.UserOrgSession, _ orgkey.OrgKey, _ ...pscrbac.Permission) (bool, error) {
	return true, nil
}

func NewRBACMiddleware(demandService demandChecker, permissions ...pscrbac.Permission) RbacMiddleWare {
	return RbacMiddleWare{demandService: demandService, permissions: permissions}
}

func getRealDemandService(cfg *config.Config) demandChecker {
	demandClient := pscrbac.NewDemandService(pschttp.NewClient(
		pschttp.WithUserAgent(cfg.UserAgent),
		pschttp.WithPlugin(
			timeout.Request(cfg.RbacHTTPHandleTimeout),
			plugin.NewErrorPlugin(LoggerHandlerError),
			plugin.NewResponsePlugin(LoggerHandlerResponse))))
	return &demandClient
}

func getLocalDemandService() demandChecker {
	return localDemandChecker{}
}

//goland:noinspection GoMixedReceiverTypes
func (mwf RouteMiddlewareFactory) RBACMiddleWare(permissions ...pscrbac.Permission) gin.HandlerFunc {
	var demand demandChecker
	if mwf.cfg.IsLocalEnv() || mwf.cfg.IsDevEnv() {
		demand = getLocalDemandService()
	} else {
		demand = getRealDemandService(mwf.cfg)
	}
	return NewRBACMiddleware(demand, permissions...).getHandler()
}

func (rbac RbacMiddleWare) getHandler() gin.HandlerFunc {
	return func(c *gin.Context) {
		userContext, foundUserContext := c.Get(pschttp.UserContextHeader)
		if !foundUserContext {
			c.AbortWithStatus(http.StatusUnauthorized)
			return
		}

		requestContext, foundRequestContext := c.Get(pschttp.RequestContextHeader)
		if !foundRequestContext {
			c.AbortWithStatus(http.StatusUnauthorized)
			return
		}

		orgKey := orgkey.OrgKey{}
		orgKeyString := c.Param("org_key")
		if orgKeyString != "" {
			orgKeyTemp, err := orgkey.Decode(orgKeyString)
			if err != nil {
				c.AbortWithStatus(http.StatusUnauthorized)
				return
			}
			orgKey = *orgKeyTemp
		}

		userAuth := pschttp.NewImpersonatedAuth(requestContext.(pschttp.RequestContext), userContext.(pschttp.UserContext))
		hasAccess, err := rbac.demandService.HasPermission(c.Request.Context(), userAuth,
			userContext.(pschttp.UserContext).LoginSession, orgKey, rbac.permissions...)
		if err != nil {
			if errors.Is(err, context.DeadlineExceeded) {
				_ = c.AbortWithError(http.StatusRequestTimeout, err)
				return
			}
			_ = c.AbortWithError(http.StatusServiceUnavailable, err)
			return
		}
		if !hasAccess {
			c.AbortWithStatus(http.StatusForbidden)
			return
		}
		c.Next()
	}
}

func LoggerHandlerError(ctx *gcontext.Context, handler gcontext.Handler) {
	log := logger.RequestLoggerFromContext(ctx.Request.Context())
	log.Error("error from demand service request", zap.String("uri", ctx.Request.RequestURI), zap.NamedError("demand_service_error", ctx.Error))
	handler.Next(ctx)
}

func LoggerHandlerResponse(ctx *gcontext.Context, handler gcontext.Handler) {
	log := logger.RequestLoggerFromContext(ctx.Request.Context())
	log.Debug("response from demand service request", zap.String("uri", ctx.Request.RequestURI), zap.Int("status", ctx.Response.StatusCode))
	handler.Next(ctx)
}
