package middleware

import (
	"net/http"
	"strconv"
	"strings"

	"github.com/felixge/httpsnoop"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
)

const (
	// Common Labels
	labelMethod        = "method"
	labelCode          = "code"
	labelOperationName = "operation_name"

	// Requests Total Metric Name and Labels
	metricNameRequestsTotal = "requests_total"

	// Request Duration Metric Name and Labels
	metricNameRequestsDurationSeconds      = "request_duration_seconds"
	metricNameRequestsDurationSecondsGauge = "request_duration_seconds_gauge"
)

type Metrics struct {
	requestsReceived     *prometheus.CounterVec
	requestDuration      *prometheus.HistogramVec
	requestDurationGauge *prometheus.GaugeVec
}

func RegisterPromMetrics() *Metrics {
	reqReceived := prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Namespace: "http",
			Name:      metricNameRequestsTotal,
			Help:      "Total number of requests received.",
		},
		[]string{labelOperationName, labelMethod, labelCode},
	)
	prometheus.MustRegister(reqReceived)

	reqDuration := prometheus.NewHistogramVec(
		prometheus.HistogramOpts{
			Namespace: "http",
			Name:      metricNameRequestsDurationSeconds,
			Help:      "Duration of a request in seconds.",
		},
		[]string{labelOperationName, labelMethod, labelCode},
	)
	prometheus.MustRegister(reqDuration)

	// Because APM doesn't support histograms, we will also track request duration as a gauge
	reqDurationGauge := prometheus.NewGaugeVec(prometheus.GaugeOpts{
		Namespace: "http",
		Name:      metricNameRequestsDurationSecondsGauge,
		Help:      "Duration of a request in seconds, as a gauge.",
	},
		[]string{labelOperationName, labelMethod, labelCode},
	)
	prometheus.MustRegister(reqDurationGauge)

	return &Metrics{
		requestsReceived:     reqReceived,
		requestDuration:      reqDuration,
		requestDurationGauge: reqDurationGauge,
	}
}

func (m *Metrics) Default() http.Handler {
	return promhttp.Handler()
}

func (m *Metrics) IncRequests(operationName string, method string, code int) {
	m.requestsReceived.WithLabelValues(operationName, method, strconv.Itoa(code)).Inc()
}

func (m *Metrics) ObsDuration(operationName string, method string, code int, duration float64) {
	m.requestDuration.WithLabelValues(operationName, method, strconv.Itoa(code)).Observe(duration)
	m.requestDurationGauge.WithLabelValues(operationName, method, strconv.Itoa(code)).Set(duration)
}

func (mwf RouteMiddlewareFactory) REDMetrics(promMetrics *Metrics, next http.HandlerFunc) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		m := httpsnoop.CaptureMetrics(next, w, r)
		path := r.URL.Path

		if strings.Contains(path, "health_check") {
			return
		}

		promMetrics.IncRequests(operationNameFunc(r), r.Method, m.Code)
		promMetrics.ObsDuration(operationNameFunc(r), r.Method, m.Code, m.Duration.Seconds())
	}
}

func operationNameFunc(r *http.Request) string {
	splitPath := strings.Split(r.URL.Path, "/")
	method := r.Method
	var operationName string = "UKNOWN"
	switch method {
	case "PUT":
		operationName = strings.TrimSuffix(splitPath[len(splitPath)-2], "s")
	case "POST":
		operationName = splitPath[len(splitPath)-1]
	}
	return "HTTP " + r.Method + " " + operationName
}
