// Copyright (c) 2023 VMware Carbon Black, all rights reserved

// Package models -- thos package contains the various data structures used to configure the hint generator
package models

import (
	"bytes"
	"fmt"
	"net/http"
	"net/http/httptest"
	"time"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicechangetype"
	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

func yesterday() time.Time {
	now := time.Now()
	oneDay, _ := time.ParseDuration("24h")
	yesterday := now.Add(-oneDay)
	return yesterday
}

var _ = Describe("Device Model validation tests", func() {
	orgKeyObject, _ := orgkeylib.Encode("enki", 1337)
	orgKey := orgKeyObject.String()

	// ----- URL Bind testing --------------------
	// NOTE: These handle supplied url API content validation, not value validation

	sampleDeviceBody := `{"changes":["DEV_GROUP_CHANGE", "DEV_SENSOR_KIT_UPDATE"]}`

	// NOTE: Same effect as supplying empty string orgKey
	It("Ensure we reject missing orgKey", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/device/uri", bytes.NewBufferString(sampleDeviceBody))
		context.Params = []gin.Param{
			{Key: "deviceId", Value: "1337"},
		}
		apiDeviceModel := APIDevice{}
		err := context.ShouldBindUri(apiDeviceModel.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIDeviceURI.OrgKey' Error:Field validation for 'OrgKey' failed on the 'required' tag"))
	})

	// NOTE: Same effect as supplying empty string deviceId
	It("Ensure we reject missing deviceId", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/device/uri", bytes.NewBufferString(sampleDeviceBody))
		context.Params = []gin.Param{
			{Key: "orgKey", Value: orgKey},
		}
		apiDeviceModel := APIDevice{}
		err := context.ShouldBindUri(apiDeviceModel.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIDeviceURI.Id' Error:Field validation for 'Id' failed on the 'required' tag"))
	})

	It("Ensure we reject a non-numeric deviceId", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/device/uri", bytes.NewBufferString(sampleDeviceBody))
		context.Params = []gin.Param{
			{Key: "orgKey", Value: orgKey},
			{Key: "deviceId", Value: "BOGUS"},
		}
		apiDeviceModel := APIDevice{}
		err := context.ShouldBindUri(apiDeviceModel.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("strconv.ParseInt: parsing \"BOGUS\": invalid syntax"))
	})

	// ----- Body Bind testing --------------------
	// NOTE: These handle supplied body API content validation, not value validation

	It("Ensure we can parse a filled in device body", func() {
		rawGroupString := `{"changes":["DEV_GROUP_CHANGE", "DEV_SENSOR_KIT_UPDATE"], "live_query_time":1672531200000}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/device/uri", bytes.NewBufferString(rawGroupString))

		apiDeviceModel := APIDevice{}
		err := context.ShouldBindJSON(apiDeviceModel.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiDeviceModel.Changes)).To(BeEquivalentTo(2))
		Expect(*apiDeviceModel.LiveQueryTime).To(BeEquivalentTo(1672531200000))
	})

	It("Ensure we allow an empty JSON body", func() {
		rawGroupString := `{}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawGroupString))

		apiDeviceModel := APIDevice{}
		err := context.ShouldBindJSON(apiDeviceModel.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiDeviceModel.Changes)).To(BeEquivalentTo(0))
		Expect(apiDeviceModel.LiveQueryTime).To(BeNil())
	})

	It("Ensure we reject an empty string body", func() {
		rawGroupString := ""

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawGroupString))

		apiDeviceModel := APIDevice{}
		err := context.ShouldBindJSON(apiDeviceModel.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("EOF")) // NOTE: best we can do with gomega
	})

	// -----

	It("Ensure we reject non-list changes", func() {
		rawGroupString := `{"changes": "DEV_SENSOR_CERTIFICATE_REQUEST"}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawGroupString))

		apiDeviceModel := APIDevice{}
		err := context.ShouldBindJSON(apiDeviceModel.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("cannot unmarshal string into Go struct field APIDeviceBody.changes of type []string"))
	})

	// NOTE:
	It("Ensure we allow missing changes", func() {
		rawGroupString := `{"live_query_time":1672531200000}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/group/uri", bytes.NewBufferString(rawGroupString))

		apiDeviceModel := APIDevice{}
		err := context.ShouldBindJSON(apiDeviceModel.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiDeviceModel.Changes)).To(BeEquivalentTo(0))
		Expect(*apiDeviceModel.LiveQueryTime).To(BeEquivalentTo(1672531200000))
	})

	It("Ensure we allow empty changes list", func() {
		rawGroupString := `{"device_type":"WINDOWS", "deployment_type":"ENDPOINT", "changes":[]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/device/uri", bytes.NewBufferString(rawGroupString))

		apiDeviceModel := APIDevice{}
		err := context.ShouldBindJSON(apiDeviceModel.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiDeviceModel.Changes)).To(BeEquivalentTo(0))
	})

	It("Ensure we reject excess changes", func() {
		rawGroupString := `{"changes":["DEV_GROUP_CHANGE", "DEV_SENSOR_CONFIG_UPDATE", "DEV_SENSOR_CERTIFICATE_REQUEST", ` +
			`"DEV_HASH_DELETE_REQUEST", "DEV_HASH_SAMPLE_REQUEST", "DEV_SENSOR_LOGS_REQUEST", "DEV_SENSOR_KIT_UPDATE", ` +
			`"DEV_SENSOR_TRIGGER_PERFMON", "DEV_SENSOR_ACTION", "DEV_REREGISTER", "DEV_LIVE_QUERY_CHANGE", "DEV_GROUP_CHANGE", "DEV_REP_REQUEST"]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/device/uri", bytes.NewBufferString(rawGroupString))

		apiDeviceModel := APIDevice{}
		err := context.ShouldBindJSON(apiDeviceModel.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIDeviceBody.Changes' Error:Field validation for 'Changes' failed on the 'max' tag"))
	})

	// ----- DB Conversion testing --------------------
	// NOTE: These validate DB structure conversion checks

	useAPI := APIDeviceURI{Id: 1337, OrgKey: orgKey}
	changeTime := int64(1672531200000)

	It("Ensure we convert to a valid device DB object", func() {
		apiDevice := APIDevice{
			APIDeviceURI: useAPI,
			APIDeviceBody: APIDeviceBody{
				Changes:       devicechangetype.AllDeviceChangeTypeArray,
				LiveQueryTime: &changeTime,
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())

		Expect(device).ToNot(BeNil())
		Expect(device.Id).To(BeEquivalentTo(1337))
		Expect(device.OrgKey).To(BeEquivalentTo(orgKey))
		Expect(device.SensorKitUpdateTime).ToNot(BeNil())
		Expect(device.GroupChangeTime).ToNot(BeNil())
		Expect(device.SensorConfigUpdateTime).ToNot(BeNil())
		Expect(device.SensorCertificateUpdateTime).ToNot(BeNil())
		Expect(device.HashDeleteRequestTime).ToNot(BeNil())
		Expect(device.HashUploadRequestTime).ToNot(BeNil())
		Expect(device.LogsRequestTime).ToNot(BeNil())
		Expect(device.SensorKitUpdateTime).ToNot(BeNil())
		Expect(device.PerfmonTriggerTime).ToNot(BeNil())
		Expect(device.SensorActionTime).ToNot(BeNil())
		Expect(device.ReRegisterTime).ToNot(BeNil())
		Expect(device.LiveQueryChangeTime).ToNot(BeNil())

		Expect(device.DeploymentType).To(BeNil()) // shouild be filled in!
		Expect(device.DeviceType).To(BeNil())     // shouild be filled in!
	})

	// ----- URI

	badAPI := APIDeviceURI{Id: 1337, OrgKey: "BOGUS123"}

	It("Ensure we catch an invalid orgkey", func() {
		apiDevice := APIDevice{
			APIDeviceURI: badAPI,
			APIDeviceBody: APIDeviceBody{
				Changes: devicechangetype.AllDeviceChangeTypeArray,
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(device).To(BeNil())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("error decoding org key BOGUS123: alphabet used for hash was different"))
	})

	badAPI2 := APIDeviceURI{Id: 1337, OrgKey: "     "}

	It("Ensure we catch an empty space orgkey", func() {
		apiDevice := APIDevice{
			APIDeviceURI: badAPI2,
			APIDeviceBody: APIDeviceBody{
				Changes: devicechangetype.AllDeviceChangeTypeArray,
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(device).To(BeNil())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("org key is required"))
	})

	badAPI3 := APIDeviceURI{Id: 1337, OrgKey: "  " + orgKey + "\n"}

	It("Ensure we allow an orgkey with leading/trailing whitespace (with warnings)", func() {
		apiDevice := APIDevice{
			APIDeviceURI: badAPI3,
			APIDeviceBody: APIDeviceBody{
				Changes: devicechangetype.AllDeviceChangeTypeArray,
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(err).To(MatchError(ErrModelConversionWarning))
		Expect(device).ToNot(BeNil())
		Expect(device.OrgKey).To(BeEquivalentTo(orgKey))
	})

	// ----- Body

	It("Ensure we reject an invalid DeviceChangeType for Changes", func() {
		apiDevice := APIDevice{
			APIDeviceURI: useAPI,
			APIDeviceBody: APIDeviceBody{
				Changes: []string{"BOGUS_DEVICE_CHANGE_TYPE"},
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(err).To(MatchError(ErrModelConversionWarning))
		Expect(device).ToNot(BeNil())
	})

	// NOTE: Duplicate changes tested in functional tests (requires live service)

	// ----- Added for live_query_time checking

	It("Ensure we can provide a timestamp for a livequery change", func() {
		yesterday := yesterday().UnixMilli()
		apiDevice := APIDevice{
			APIDeviceURI: useAPI,
			APIDeviceBody: APIDeviceBody{
				Changes:       []string{devicechangetype.DEV_LIVE_QUERY_CHANGE.String()},
				LiveQueryTime: &yesterday,
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(device).ToNot(BeNil())
		Expect(device.LiveQueryChangeTime).ToNot(BeNil())
		Expect(device.LiveQueryChangeTime.UnixMilli()).To(BeEquivalentTo(yesterday))
	})

	It("Ensure that a provided timestamp does not affect other change types", func() {
		yesterday := yesterday().UnixMilli()
		apiDevice := APIDevice{
			APIDeviceURI: useAPI,
			APIDeviceBody: APIDeviceBody{
				Changes:       []string{devicechangetype.DEV_GROUP_CHANGE.String()},
				LiveQueryTime: &yesterday,
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(device).ToNot(BeNil())
		Expect(device.LiveQueryChangeTime).To(BeNil())
		Expect(device.GroupChangeTime.UnixMilli()).To(BeNumerically("~", time.Now().UnixMilli(), 5))
	})

	It("Ensure livequery change with no timestamp uses current time", func() {
		apiDevice := APIDevice{
			APIDeviceURI: useAPI,
			APIDeviceBody: APIDeviceBody{
				Changes: []string{devicechangetype.DEV_LIVE_QUERY_CHANGE.String()},
			},
		}
		device, err := apiDevice.ConvertToDBObject()
		Expect(err).Should(MatchError(ErrModelConversionWarning))
		Expect(device).ToNot(BeNil())
		Expect(device.LiveQueryChangeTime).ToNot(BeNil())
		Expect(device.LiveQueryChangeTime.UnixMilli()).To(BeNumerically("~", time.Now().UnixMilli(), 5))
	})
})
