// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package models

import (
	"bytes"
	"fmt"
	"net/http"
	"net/http/httptest"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

var _ = Describe("API Organization validation tests", func() {
	orgKeyObject, _ := orgkeylib.Encode("org", 1337)
	orgKey := orgKeyObject.String()

	// ----- URL Bind testing --------------------
	// NOTE: These handle supplied url API content validation, not value validation

	It("Ensure we require orgKey", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(""))
		context.Params = []gin.Param{}
		apiOrg := APIOrg{}
		err := context.ShouldBindUri(apiOrg.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIOrgURI.OrgKey' Error:Field validation for 'OrgKey' failed on the 'required' tag"))
	})

	It("Ensure orgKey cant be all whitespace", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString("         "))
		context.Params = []gin.Param{}
		apiOrg := APIOrg{}
		err := context.ShouldBindUri(apiOrg.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIOrgURI.OrgKey' Error:Field validation for 'OrgKey' failed on the 'required' tag"))
	})

	// ----- Body Bind testing --------------------
	// NOTE: These handle supplied body API content validation, not value validation

	It("Ensure we can parse a valid org structure", func() {
		rawOrgString := `{"changes":["ORG_DISABLE_DEFENSE_RULES"],
						"deployment_type_changes":{"ENDPOINT": ["ORG_HASH_LIST_CHANGE"]},
						"is_kill_switch_on":true,
						"canary_payload_device_types":["WINDOWS", "IOS", "LINUX"]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())

		Expect(*apiOrg.IsKillSwitchOn).To(BeEquivalentTo(true))
		Expect(len(apiOrg.Changes)).To(BeEquivalentTo(1))
		Expect(len(apiOrg.DeploymentTypeChanges)).To(BeEquivalentTo(1))
		Expect(len(apiOrg.DeploymentTypeChanges[string(deploymenttype.ENDPOINT)])).To(BeEquivalentTo(1))
		Expect(len(apiOrg.CanaryPayloadDeviceTypes)).To(BeEquivalentTo(3))
	})

	It("Ensure we can parse a valid org structure with is_kill_switch_on false", func() {
		rawOrgString := `{"changes":["ORG_DISABLE_DEFENSE_RULES"],
						"deployment_type_changes":{"ENDPOINT": ["ORG_HASH_LIST_CHANGE"]},
						"is_kill_switch_on":false,
						"canary_payload_device_types":["WINDOWS", "IOS", "LINUX"]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())

		Expect(*apiOrg.IsKillSwitchOn).To(BeEquivalentTo(false))
		Expect(len(apiOrg.Changes)).To(BeEquivalentTo(1))
		Expect(len(apiOrg.DeploymentTypeChanges)).To(BeEquivalentTo(1))
		Expect(len(apiOrg.DeploymentTypeChanges[string(deploymenttype.ENDPOINT)])).To(BeEquivalentTo(1))
		Expect(len(apiOrg.CanaryPayloadDeviceTypes)).To(BeEquivalentTo(3))
	})

	It("Ensure we reject if empty string body specified", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(""))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("EOF")) // NOTE: best we can do with gomega
	})

	It("Ensure we accept org with empty json body (no changes)", func() {
		// See DSER-49110
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString("{}"))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(apiOrg.IsKillSwitchOn).To(BeNil())
		Expect(len(apiOrg.Changes)).To(BeEquivalentTo(0))
		Expect(len(apiOrg.DeploymentTypeChanges)).To(BeEquivalentTo(0))
		Expect(len(apiOrg.CanaryPayloadDeviceTypes)).To(BeEquivalentTo(0))
	})

	// -----

	It("Ensure we reject non-list changes", func() {
		rawOrgString := `{"changes":"ORG_HASH_LIST_CHANGE"}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("cannot unmarshal string into Go struct field APIOrgBody.changes of type []string"))
	})

	It("Ensure we accept empty changes list", func() {
		// See DSER-49110

		rawOrgString := `{"is_kill_switch_on":true, "changes":[]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiOrg.Changes)).To(BeEquivalentTo(0))
	})

	It("Ensure we reject excess changes", func() {
		rawOrgString := `{"changes":["ORG_NET_CFG_CHANGE",  "ORG_HASH_LIST_CHANGE", "ORG_AV_INTEGRATION_CHANGE", ` +
			`"ORG_KILL_SWITCH_ON", "ORG_UNINSTALL_CODE_CHANGE", "ORG_LIVE_QUERY_CHANGE", "ORG_SET_MELTDOWN_REGISTRY_KEY", ` +
			`"ORG_THREATHUNTER_RULES", "ORG_DISABLE_DEFENSE_RULES", "ORG_BLADE_ACTION_CHANGE", "ORG_NET_CFG_CHANGE"]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIOrgBody.Changes' Error:Field validation for 'Changes' failed on the 'max' tag"))
	})

	It("Ensure we reject non-map deployment_type_changes", func() {
		rawOrgString := `{"deployment_type_changes":["ENDPOINT"]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("cannot unmarshal array into Go struct field APIOrgBody.deployment_type_changes of type map[string][]string"))
	})

	It("Ensure we accept empty deployment_type_changes", func() {
		// See DSER-49110

		rawOrgString := `{"deployment_type_changes":{}}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiOrg.DeploymentTypeChanges)).To(BeEquivalentTo(0))
	})

	It("Ensure we reject excess deployment_type_changes", func() {
		// NOTE: added bogus entry at end to allow for max size checking
		dtc := `"ENDPOINT": ["ORG_HASH_LIST_CHANGE"], "WORKLOAD": ["ORG_HASH_LIST_CHANGE"], "VDI": ["ORG_HASH_LIST_CHANGE"], ` +
			`"AWS": ["ORG_HASH_LIST_CHANGE"], "AZURE": ["ORG_HASH_LIST_CHANGE"], "GCP": ["ORG_HASH_LIST_CHANGE"], "BOGUS": ["ORG_HASH_LIST_CHANGE"]`
		rawOrgString := `{"deployment_type_changes":{` + dtc + `}}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIOrgBody.DeploymentTypeChanges' Error:Field validation for 'DeploymentTypeChanges' failed on the 'max' tag"))
	})

	It("Ensure we reject non-list deployment_type_changes values", func() {
		dtc := `"ENDPOINT": "ORG_THREATHUNTER_RULES"`
		rawOrgString := `{"deployment_type_changes":{` + dtc + `}}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("cannot unmarshal string into Go struct field APIOrgBody.deployment_type_changes of type []string"))
	})

	It("Ensure we accept empty deployment_type_changes entry values", func() {
		// See DSER-49110

		dtc := `"ENDPOINT": []`
		rawOrgString := `{"deployment_type_changes":{` + dtc + `}}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiOrg.DeploymentTypeChanges)).To(BeEquivalentTo(1))
		Expect(len(apiOrg.DeploymentTypeChanges[string(deploymenttype.ENDPOINT)])).To(BeEquivalentTo(0))
	})

	It("Ensure we reject excess deployment_type_changes entry values", func() {
		dtc := `"ENDPOINT": ["ORG_THREATHUNTER_RULES", "ORG_NET_CFG_CHANGE", "ORG_HASH_LIST_CHANGE", "ORG_AV_INTEGRATION_CHANGE", ` +
			`"ORG_KILL_SWITCH_ON", "ORG_UNINSTALL_CODE_CHANGE", "ORG_LIVE_QUERY_CHANGE", "ORG_SET_MELTDOWN_REGISTRY_KEY", ` +
			`"ORG_DISABLE_DEFENSE_RULES", "ORG_BLADE_ACTION_CHANGE", "ORG_THREATHUNTER_RULES"]`
		rawOrgString := `{"deployment_type_changes":{` + dtc + `}}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIOrgBody.DeploymentTypeChanges[ENDPOINT]' Error:Field validation for 'DeploymentTypeChanges[ENDPOINT]' failed on the 'max' tag"))
	})

	It("Ensure is_kill_switch_on can be specified alone", func() {
		rawOrgString := `{"is_kill_switch_on":true}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(*apiOrg.IsKillSwitchOn).To(BeEquivalentTo(true))
	})

	It("Ensure is_kill_switch_on can be specified alone with false value", func() {
		rawOrgString := `{"is_kill_switch_on":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(*apiOrg.IsKillSwitchOn).To(BeEquivalentTo(false))
	})

	It("Ensure we reject non-list canary_payload_device_types", func() {
		rawOrgString := `{"canary_payload_device_types":"WINDOWS"}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("cannot unmarshal string into Go struct field APIOrgBody.canary_payload_device_types of type []string"))
	})

	It("Ensure we accept empty canary_payload_device_types", func() {
		// See DSER-49110

		rawOrgString := `{"canary_payload_device_types":[]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiOrg.CanaryPayloadDeviceTypes)).To(BeEquivalentTo(0))
	})

	It("Ensure we reject excess canary_payload_device_types", func() {
		rawOrgString := `{"canary_payload_device_types":["IOS", "WINDOWS", "ANDROID", "MAC", "LINUX", "OTHER", "IOS"]}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIOrgBody.CanaryPayloadDeviceTypes' Error:Field validation for 'CanaryPayloadDeviceTypes' failed on the 'max' tag"))
	})

	It("Ensure we accept no changes specified", func() {
		// See DSER-49110

		rawOrgString := `{"changes":[], "canary_payload_device_types": [], "deployment_type_changes": {}}`

		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/org/uri", bytes.NewBufferString(rawOrgString))

		apiOrg := APIOrg{}
		err := context.ShouldBindJSON(apiOrg.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
	})

	// ----- DB Conversion testing --------------------
	// NOTE: These validate DB structure conversion checks

	useAPI := APIOrgURI{OrgKey: orgKey}

	It("Ensure we convert a valid org structure", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())

		Expect(org).ToNot(BeNil())
		Expect(len(org.ChangeTimestamps)).To(BeEquivalentTo(1))
		Expect(org.IsKillSwitchOn).ToNot(BeNil())
		Expect(len(org.CanaryPayloadTimestamps)).To(BeEquivalentTo(3))
		Expect(len(org.DeploymentTypeChangeTimestamps)).To(BeEquivalentTo(1))
		Expect(org.GetEnv()).To(BeEquivalentTo("org"))
	})

	badAPI := APIOrgURI{OrgKey: "BOGUS789"}

	It("Ensure we catch an invalid orgkey", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: badAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(org).To(BeNil())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("error decoding org key BOGUS789: mismatch between encode and decode: BOGUS789 start 627RJ79Z re-encoded"))
	})

	badAPI2 := APIOrgURI{OrgKey: "    "}

	It("Ensure we catch an emptystring orgkey", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: badAPI2,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(org).To(BeNil())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("org key is required"))
	})

	badAPI3 := APIOrgURI{OrgKey: "\t" + orgKey + "\n"}

	It("Ensure we allow an orgkey with leading/trailing whitespace (with warnings, will be trimmed)", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: badAPI3,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).To(BeNil())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(org.OrgKey).To(BeEquivalentTo(orgKey))
	})

	// -----

	It("Ensure we reject invalid incoming Changes", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{"BOGUS_ORG_CHANGE"},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(org).ToNot(BeNil())
		Expect(err).To(MatchError(ErrModelConversionWarning))
	})

	// NOTE: Duplicate changes tested in functional tests (requires live service)

	It("Ensure we reject invalid incoming CanaryPayloadDeviceTypes", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"BOGUS_DEVICE_TYPE"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(org).ToNot(BeNil())
		Expect(err).To(MatchError(ErrModelConversionWarning))
	})

	// NOTE: Duplicate CanaryPayloadDeviceTypes tested in functional tests (requires live service)

	It("Ensure we reject invalid incoming DeploymentTypeChanges", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{"BOGUS_DEPLOYMENT_TYPE": {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(org).ToNot(BeNil())
		Expect(err).To(MatchError(ErrModelConversionWarning))
	})

	// NOTE: Duplicate DeploymentTypeChanges tested in functional tests (requires live service)

	It("Ensure we reject invalid incoming DeploymentTypeChanges entry", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {"BOGUS_ORG_CHANGE"}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(org).ToNot(BeNil())
		Expect(err).To(MatchError(ErrModelConversionWarning))
	})
	// NOTE: Duplicate DeploymentTypeChanges entry tested in functional tests (requires live service)

	It("Ensure that live_query_time supplied is used for the DeploymentTypeChanges when ORG_LIVE_QUERY_CHANGE is used", func() {
		yesterday := yesterday().UnixMilli()

		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {"ORG_LIVE_QUERY_CHANGE"}},
				LiveQueryTime:            &yesterday,
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(org).ToNot(BeNil())
		got := org.DeploymentTypeChangeTimestamps[deploymenttype.ENDPOINT][orgchangetype.ORG_LIVE_QUERY_CHANGE].UnixMilli()
		Expect(got).To(BeEquivalentTo(yesterday))
	})

	It("Ensure when we don't provide a timestamp for a DeploymentTypeChanges change (should default to now)", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {"ORG_LIVE_QUERY_CHANGE"}},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).To(MatchError(ErrModelConversionWarning))
		Expect(org).ToNot(BeNil())
		Expect(org.DeploymentTypeChangeTimestamps[deploymenttype.ENDPOINT][orgchangetype.ORG_LIVE_QUERY_CHANGE].Unix()).
			To(BeNumerically("~", time.Now().Unix(), 5))
	})

	It("Ensure when a timestamp for a DeploymentTypeChanges does not affect other change types (should default to now)", func() {
		yesterday := yesterday().UnixMilli()

		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {"ORG_DISABLE_DEFENSE_RULES"}},
				LiveQueryTime:            &yesterday,
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(org).ToNot(BeNil())
		Expect(org.DeploymentTypeChangeTimestamps[deploymenttype.ENDPOINT][orgchangetype.ORG_DISABLE_DEFENSE_RULES].Unix()).
			To(BeNumerically("~", time.Now().Unix(), 5))
	})

	It("Ensure we use the provided LiveQueryTime when ORG_LIVE_QUERY_CHANGE is provided in changes", func() {
		yesterday := yesterday().UnixMilli()

		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_LIVE_QUERY_CHANGE.String()},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				LiveQueryTime:            &yesterday,
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(org).ToNot(BeNil())
		Expect(org.ChangeTimestamps[orgchangetype.ORG_LIVE_QUERY_CHANGE].UnixMilli()).To(BeEquivalentTo(yesterday))
	})

	It("Ensure we use the current time when LiveQueryTime is not provided and ORG_LIVE_QUERY_CHANGE provided in changes", func() {
		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String(), "ORG_LIVE_QUERY_CHANGE"},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).To(MatchError(ErrModelConversionWarning))
		Expect(org).ToNot(BeNil())
		Expect(org.ChangeTimestamps[orgchangetype.ORG_LIVE_QUERY_CHANGE].Unix()).To(BeNumerically("~", time.Now().Unix(), 5))
	})

	It("Ensure when we specify LiveQueryTime it does not affect other change types", func() {
		yesterday := yesterday().UnixMilli()

		isKillSwitchOn := true
		apiOrg := APIOrg{
			APIOrgURI: useAPI,
			APIOrgBody: APIOrgBody{
				Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String(), "ORG_LIVE_QUERY_CHANGE"},
				IsKillSwitchOn:           &isKillSwitchOn,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
				LiveQueryTime:            &yesterday,
			},
		}
		org, err := apiOrg.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())
		Expect(org).ToNot(BeNil())
		Expect(org.ChangeTimestamps[orgchangetype.ORG_DISABLE_DEFENSE_RULES].Unix()).To(BeNumerically("~", time.Now().Unix(), 5))
	})
})
