// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package models

import (
	"errors"
	"fmt"
	"strings"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"

	"gitlab.bit9.local/psc/ems/hinting/internal/utils"

	"gorm.io/gorm"
	"gorm.io/gorm/clause"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/types"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"

	"go.uber.org/zap/zapcore"
)

////////////////////////////////////////////////////////////////////////////////
// Postgres Database Specifics

// Environment -- environment structure stored in postgres
type Environment struct {
	Name                    string `gorm:"primaryKey"`
	CertServiceTime         *time.Time
	PolicyServiceUpdateTime *time.Time
	CanaryPayloadTimestamps types.DeviceTypeToTimeMap `gorm:"type:JSONB"`
}

func (env *Environment) CacheKey() string {
	return fmt.Sprintf("environments-%s", env.Name)
}

func (env *Environment) GetOnConflictClause() clause.OnConflict {
	return clause.OnConflict{
		Columns:   env.GetPrimaryColumnNames(),
		DoUpdates: env.GetUpsertClauses(),
	}
}

func (env *Environment) GetPrimaryColumnNames() []clause.Column {
	return []clause.Column{{Name: "name"}}
}

func (env *Environment) GetUpsertClauses() clause.Set {
	gormExprMap := getGreatestExpr("environments", "cert_service_time", "policy_service_update_time")
	gormExprMap["canary_payload_timestamps"] = gorm.Expr("jsonb_recursive_merge(environments.canary_payload_timestamps,excluded.canary_payload_timestamps)")
	return clause.Assignments(gormExprMap)
}

func (env *Environment) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	encoder.AddString("env", env.Name)
	utils.AddTimePointerToEncoder(encoder, "cert_service_time", env.CertServiceTime)
	utils.AddTimePointerToEncoder(encoder, "policy_service_update_time", env.PolicyServiceUpdateTime)
	var err error
	if len(env.CanaryPayloadTimestamps) > 0 {
		err = encoder.AddObject("canary_payload_timestamps", env.CanaryPayloadTimestamps)
	}
	return err
}

func (env *Environment) GetMetricDimensions() *metrics.Dimensions {
	if env.CertServiceTime == nil && env.PolicyServiceUpdateTime == nil && len(env.CanaryPayloadTimestamps) == 0 {
		return nil
	}

	var hints []string
	dimensions := metrics.Dimensions{
		metrics.DimensionEnvName: env.Name,
	}

	if env.CertServiceTime != nil {
		hints = append(hints, "env_CertServiceTime")
	}
	if env.PolicyServiceUpdateTime != nil {
		hints = append(hints, "env_PolicyServiceUpdateTime")
	}

	if len(hints) > 0 {
		dimensions["env_hints"] = strings.Join(hints, ",")
	}

	var dtHints []string
	for deviceType := range env.CanaryPayloadTimestamps {
		dtHints = append(dtHints, string(deviceType))
	}
	if len(dtHints) > 0 {
		dimensions["env_canary_payload"] = strings.Join(dtHints, ",")
	}

	return &dimensions
}

//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// API Input Handling

type APIEnvURI struct {
	Name string `uri:"envName" binding:"required"`
}

type APIEnvBody struct {
	// Indicates if CertServiceTime should be updated.
	//
	// required: false
	CertServiceTimeChanged bool `json:"cert_service_time_changed"`

	// Indicates if PolicyServiceUpdateTime should be updated.
	//
	// required: false
	PolicyServiceTimeChanged bool `json:"policy_service_time_changed"`

	// List of device types to be updated.
	//
	// Valid Device Types:
	// 	ANDROID,
	//	IOS,
	//	LINUX,
	//	MAC,
	//	OTHER,
	//	WINDOWS
	//
	// required: false
	// example: ["WINDOWS", "MAC", "LINUX"]
	CanaryPayloadDeviceTypes []string `json:"canary_payload_device_types" binding:"omitempty,max=6"`
}

type APIEnv struct {
	APIEnvURI
	APIEnvBody
}

// GetURIPart -- Returns the APIEnvURI component of an APIEnv
func (apiEnv *APIEnv) GetURIPart() URIPart[*Environment] {
	return &apiEnv.APIEnvURI
}

// GetBodyPart -- Returns the APIEnvBody component of an APIEnv
func (apiEnv *APIEnv) GetBodyPart() any {
	return &apiEnv.APIEnvBody
}

// ConvertToDBObject -- Initialize an Env object with a received APIEnvURI
func (uri *APIEnvURI) ConvertToDBObject() (*Environment, error) {
	return &Environment{Name: uri.Name, CanaryPayloadTimestamps: make(types.DeviceTypeToTimeMap)}, nil
}

// ConvertToDBObject -- Translates the APIEnv structure obtained from a user/other service into a Env structure
// that is stored in the postgres database.
func (apiEnv *APIEnv) ConvertToDBObject() (env *Environment, err error) {
	now := time.Now()
	env, _ = apiEnv.GetURIPart().ConvertToDBObject()

	env.Name = utils.Sanitize("name", env.Name)
	if env.Name == "" {
		return nil, errors.New("failed to convert Env.Name (cannot be empty string or all whitespace)")
	}

	if apiEnv.CertServiceTimeChanged {
		env.CertServiceTime = &now
	}
	if apiEnv.PolicyServiceTimeChanged {
		env.PolicyServiceUpdateTime = &now
	}

	for _, deviceTypeString := range apiEnv.CanaryPayloadDeviceTypes {
		deviceType := devicetype.FromString(deviceTypeString)
		env.CanaryPayloadTimestamps[deviceType] = now
		if !deviceType.IsKnown() {
			err = NewModelConversionWarning("unknown DeviceType '%s'", deviceTypeString)
		}
	}

	return env, err
}

//
////////////////////////////////////////////////////////////////////////////////

// NewEnv -- Create a filled in Env database object with empty CanaryPayloadTimestamps
func NewEnv(name string, certTime, policyServiceTime *time.Time) *Environment {
	return &Environment{Name: name, CertServiceTime: certTime, PolicyServiceUpdateTime: policyServiceTime, CanaryPayloadTimestamps: types.NewEmptyCanaryPayloadSet()}
}

// NewEnvWithCanaryTimes -- Create a filled in Env database object with CanaryPayloadTimestamps having all device types
func NewEnvWithCanaryTimes(name string, certTime, policyServiceTime *time.Time, allDeviceTypeCanaryTime time.Time) *Environment {
	return &Environment{Name: name, CertServiceTime: certTime, PolicyServiceUpdateTime: policyServiceTime, CanaryPayloadTimestamps: types.NewAllDeviceCanaryPayLoadSetAtTime(allDeviceTypeCanaryTime)}
}
