// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package models

import (
	"database/sql"
	"fmt"
	"strconv"
	"strings"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"

	"gorm.io/gorm"
	"gorm.io/gorm/clause"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/types"
	"gitlab.bit9.local/psc/ems/hinting/internal/utils"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"

	"go.uber.org/zap/zapcore"
)

// //////////////////////////////////////////////////////////////////////////////
// Postgres Database Specifics

// Organization -- organization structure stored in postgres
type Organization struct {
	OrgKey                         string `gorm:"primarykey" `
	IsKillSwitchOn                 sql.NullBool
	ChangeTimestamps               types.OrgChangeset               `gorm:"type:JSONB"`
	DeploymentTypeChangeTimestamps types.DeploymentTypeToChangesMap `gorm:"type:JSONB"`
	CanaryPayloadTimestamps        types.DeviceTypeToTimeMap        `gorm:"type:JSONB"`
	env                            string
}

func (org *Organization) CacheKey() string {
	return fmt.Sprintf("organizations-{%s}", org.getOrgKey())
}

func (org *Organization) getOrgKey() string {
	return org.OrgKey
}

// GetEnv -- Original environment name as determined by decoding the orgkey on migration
func (org *Organization) GetEnv() string {
	return org.env
}

func (org *Organization) GetOnConflictClause() clause.OnConflict {
	return clause.OnConflict{
		Columns:   org.GetPrimaryColumnNames(),
		DoUpdates: org.GetUpsertClauses(),
	}
}

func (org *Organization) GetPrimaryColumnNames() []clause.Column {
	return []clause.Column{{Name: "org_key"}}
}

func (org *Organization) GetUpsertClauses() clause.Set {
	gormExprMap := make(map[string]any, 5)
	gormExprMap["is_kill_switch_on"] = gorm.Expr("coalesce(excluded.is_kill_switch_on, organizations.is_kill_switch_on)")
	gormExprMap["change_timestamps"] = gorm.Expr("jsonb_recursive_merge(organizations.change_timestamps, excluded.change_timestamps)")
	gormExprMap["deployment_type_change_timestamps"] = gorm.Expr("jsonb_recursive_merge(organizations.deployment_type_change_timestamps, excluded.deployment_type_change_timestamps)")
	gormExprMap["canary_payload_timestamps"] = gorm.Expr("jsonb_recursive_merge(organizations.canary_payload_timestamps, excluded.canary_payload_timestamps)")
	return clause.Assignments(gormExprMap)
}

func (org *Organization) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	encoder.AddString("org_key", org.OrgKey)
	encoder.AddString("env", org.env)
	var retError error = nil
	if org.IsKillSwitchOn.Valid {
		encoder.AddBool("is_kill_switch_on", org.IsKillSwitchOn.Bool)
	} else {
		encoder.AddString("is_kill_switch_on", "null")
	}
	if len(org.ChangeTimestamps) > 0 {
		retError = encoder.AddObject("change_timestamps", org.ChangeTimestamps)
	}
	if len(org.DeploymentTypeChangeTimestamps) > 0 {
		retError = encoder.AddObject("deployment_type_change_timestamps", org.DeploymentTypeChangeTimestamps)
	}
	if len(org.CanaryPayloadTimestamps) > 0 {
		retError = encoder.AddObject("canary_payload_timestamps", org.CanaryPayloadTimestamps)
	}
	return retError
}

func (org *Organization) GetOrCreateDeploymentTypeChanges(deploymentType deploymenttype.DeploymentType) types.OrgChangeset {
	changes, found := org.DeploymentTypeChangeTimestamps[deploymentType]
	if found {
		return changes
	} else {
		org.DeploymentTypeChangeTimestamps[deploymentType] = types.NewEmptyChangeset()
		return org.DeploymentTypeChangeTimestamps[deploymentType]
	}
}

func (org *Organization) GetMetricDimensions() *metrics.Dimensions {
	if len(org.ChangeTimestamps) == 0 && len(org.DeploymentTypeChangeTimestamps) == 0 && len(org.CanaryPayloadTimestamps) == 0 {
		return nil
	}

	var hints []string
	dimensions := metrics.Dimensions{
		metrics.DimensionOrgKey:  org.OrgKey,
		metrics.DimensionEnvName: org.env,
		"is_killswitch_on":       strconv.FormatBool(org.IsKillSwitchOn.Valid && org.IsKillSwitchOn.Bool),
	}

	for hint := range org.ChangeTimestamps {
		hints = append(hints, hint.String())
	}
	if len(hints) > 0 {
		dimensions["org_hints"] = strings.Join(hints, ",")
	}

	for deploymentType, deploymentTypeHints := range org.DeploymentTypeChangeTimestamps {
		var dtHints []string
		for hint := range deploymentTypeHints {
			dtHints = append(dtHints, hint.String())
		}
		if len(dtHints) > 0 {
			dimensions[string(deploymentType)] = strings.Join(dtHints, ",")
		}
	}

	var dtHints []string
	for deviceType := range org.CanaryPayloadTimestamps {
		dtHints = append(dtHints, string(deviceType))
	}
	if len(dtHints) > 0 {
		dimensions["org_canary_payload"] = strings.Join(dtHints, ",")
	}

	return &dimensions
}

//
// //////////////////////////////////////////////////////////////////////////////

// //////////////////////////////////////////////////////////////////////////////
// API Input Handling

type APIOrgURI struct {
	OrgKey string `uri:"orgKey" binding:"required"`
}

type APIOrgBody struct {
	// List of organization change types to be updated.
	//
	// Valid Change Types:
	// 	ORG_AV_INTEGRATION_CHANGE,
	//	ORG_BLADE_ACTION_CHANGE,
	//	ORG_DISABLE_DEFENSE_RULES,
	//	ORG_HASH_LIST_CHANGE,
	//	ORG_KILL_SWITCH_ON,
	//	ORG_LIVE_QUERY_CHANGE,
	//	ORG_NET_CFG_CHANGE,
	//	ORG_SET_MELTDOWN_REGISTRY_KEY,
	//	ORG_THREATHUNTER_RULES,
	//	ORG_UNINSTALL_CODE_CHANGE
	//
	// required: false
	// example: ["ORG_HASH_LIST_CHANGE", "ORG_LIVE_QUERY_CHANGE"]
	Changes []string `json:"changes" binding:"omitempty,max=10"`

	// Map of deployment types, each with a list of organization change types to be updated.
	//
	// Valid Deployment Types:
	//	AWS,
	//	AZURE,
	//	ENDPOINT,
	//	GCP,
	//	VDI,
	//	WORKLOAD
	//
	// required: false
	// example: {"ENDPOINT": ["ORG_KILL_SWITCH_ON", "ORG_HASH_LIST_CHANGE"], "AWS": ["ORG_HASH_LIST_CHANGE", "ORG_THREATHUNTER_RULES"]}
	DeploymentTypeChanges map[string][]string `json:"deployment_type_changes" binding:"omitempty,max=6,dive,max=10"`

	// Flag for kill switch hint generation.
	//
	// required: false
	// default: true
	IsKillSwitchOn *bool `json:"is_kill_switch_on"`

	// List of device types to be updated.
	//
	// Valid Device Types:
	// 	ANDROID,
	//	IOS,
	//	LINUX,
	//	MAC,
	//	OTHER,
	//	WINDOWS
	//
	// required: false
	// example: ["WINDOWS", "LINUX"]
	CanaryPayloadDeviceTypes []string `json:"canary_payload_device_types" binding:"omitempty,max=6"`

	// If specified, overrides the current time value for any ORG_LIVE_QUERY_CHANGE.  Has no effect with other change types.
	//
	// required: false
	// example: 1672531200000
	LiveQueryTime *int64 `json:"live_query_time"`
}

type APIOrg struct {
	APIOrgURI
	APIOrgBody
}

// GetURIPart -- Return the APIOrgURI component of an APIOrg
func (apiOrganization *APIOrg) GetURIPart() URIPart[*Organization] {
	return &apiOrganization.APIOrgURI
}

// GetBodyPart -- Return the APIOrgBody component of an APIOrg
func (apiOrganization *APIOrg) GetBodyPart() any {
	return &apiOrganization.APIOrgBody
}

// ConvertToDBObject -- Initialize an Org object with a received APIOrg
func (uri *APIOrgURI) ConvertToDBObject() (*Organization, error) {
	uri.OrgKey = utils.Sanitize("orgkey", uri.OrgKey)
	_, env, err := utils.ValidateOrgKey(uri.OrgKey)
	if err != nil {
		return nil, err
	}
	return &Organization{OrgKey: uri.OrgKey, env: env, ChangeTimestamps: make(types.OrgChangeset), DeploymentTypeChangeTimestamps: make(types.DeploymentTypeToChangesMap), CanaryPayloadTimestamps: make(types.DeviceTypeToTimeMap)}, nil
}

// ConvertToDBObject -- Translates the APIOrg structure obtained from a user/other service into an Org structure
// that is stored in the postgres database.
func (apiOrganization *APIOrg) ConvertToDBObject() (org *Organization, err error) {
	now := time.Now()
	org, err = apiOrganization.APIOrgURI.ConvertToDBObject()
	if err != nil {
		return nil, err
	}

	if apiOrganization.IsKillSwitchOn != nil {
		org.IsKillSwitchOn = sql.NullBool{Valid: true, Bool: *apiOrganization.IsKillSwitchOn}
	} else {
		org.IsKillSwitchOn = sql.NullBool{Valid: false}
	}

	for _, change := range apiOrganization.Changes {
		changeEnum, found := orgchangetype.TranslateOrgChangeTypeString[change]
		if found {
			org.ChangeTimestamps[changeEnum] = &now

			// Special handling for ORG_LIVE_QUERY_CHANGE
			if changeEnum == orgchangetype.ORG_LIVE_QUERY_CHANGE {
				if apiOrganization.LiveQueryTime != nil {
					org.ChangeTimestamps[changeEnum] = utils.GetTimePointer(*apiOrganization.LiveQueryTime)
				} else {
					err = NewModelConversionWarning("Organization ORG_LIVE_QUERY_CHANGE did NOT receive live_query_time")
				}
			}
		} else {
			err = NewModelConversionWarning("Unknown OrgChangeType '%s'", change)
		}
	}

	for deploymentTypeString, deploymentTypeChanges := range apiOrganization.DeploymentTypeChanges {
		deploymentTypeEnum := deploymenttype.FromString(deploymentTypeString)
		if !deploymentTypeEnum.IsKnown() {
			err = NewModelConversionWarning("Unknown DeploymentType '%s'", deploymentTypeString)
		}
		for _, change := range deploymentTypeChanges {
			changeEnum, found := orgchangetype.TranslateOrgChangeTypeString[change]
			if found {
				org.GetOrCreateDeploymentTypeChanges(deploymentTypeEnum)[changeEnum] = &now
				if changeEnum == orgchangetype.ORG_LIVE_QUERY_CHANGE {
					if apiOrganization.LiveQueryTime != nil {
						org.GetOrCreateDeploymentTypeChanges(deploymentTypeEnum)[changeEnum] = utils.GetTimePointer(*apiOrganization.LiveQueryTime)
					} else {
						err = NewModelConversionWarning("Organization deployment type ORG_LIVE_QUERY_CHANGE did NOT receive live_query_time")
					}
				}
			} else {
				err = NewModelConversionWarning("Unknown OrgChangeType '%s' for DeploymentType '%s'", change, deploymentTypeString)
			}
		}
	}

	for _, deviceTypeString := range apiOrganization.CanaryPayloadDeviceTypes {
		deviceType := devicetype.FromString(deviceTypeString)
		org.CanaryPayloadTimestamps[deviceType] = now
		if !deviceType.IsKnown() {
			err = NewModelConversionWarning("Unknown DeviceType: '%s'", deviceTypeString)
		}
	}
	return org, err
}

//
// //////////////////////////////////////////////////////////////////////////////

func NewOrganizationWithoutDeploymentTypeChanges(orgKey string, killSwitchOn bool) *Organization {
	return &Organization{
		OrgKey:                         orgKey,
		IsKillSwitchOn:                 sql.NullBool{Valid: true, Bool: killSwitchOn},
		CanaryPayloadTimestamps:        make(types.DeviceTypeToTimeMap),
		ChangeTimestamps:               make(types.OrgChangeset),
		DeploymentTypeChangeTimestamps: make(types.DeploymentTypeToChangesMap),
	}
}

// NewOrg -- Create a new Org database object for testing purposes
func NewOrg(orgKey string,
	killSwitchOn bool,
	orgChangeTimestamps types.OrgChangeset,
	deploymentTypeChanges types.DeploymentTypeToChangesMap,
	deviceTypeCanaryPayloadTimestamp types.DeviceTypeToTimeMap,
) *Organization {
	return &Organization{
		OrgKey:                         orgKey,
		IsKillSwitchOn:                 sql.NullBool{Valid: true, Bool: killSwitchOn},
		ChangeTimestamps:               orgChangeTimestamps,
		CanaryPayloadTimestamps:        deviceTypeCanaryPayloadTimestamp,
		DeploymentTypeChangeTimestamps: deploymentTypeChanges,
	}
}
