// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package types

import (
	"database/sql/driver"
	"encoding/json"
	"errors"
	"fmt"
	"time"

	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"

	"gitlab.bit9.local/psc/ems/hinting/internal/logger"

	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"
)

// DeviceTypeToTimeMap -- A structure used by the Org and Env models
type DeviceTypeToTimeMap map[devicetype.DeviceType]time.Time

//goland:noinspection GoMixedReceiverTypes
func (dttm DeviceTypeToTimeMap) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	for deviceType, t := range dttm {
		encoder.AddTime(string(deviceType), t)
	}
	return nil
}

// ----- Processing methods for DeviceTypeToTimeMap --------------------

// NewEmptyCanaryPayloadSet -- Create an empty DeviceTypeToTimeMap for later use
func NewEmptyCanaryPayloadSet() DeviceTypeToTimeMap {
	return make(DeviceTypeToTimeMap)
}

// NewAllDeviceCanaryPayLoadSetAtTime -- Create a map of all DeviceTypes with the supplied time
func NewAllDeviceCanaryPayLoadSetAtTime(t time.Time) DeviceTypeToTimeMap {
	times := NewEmptyCanaryPayloadSet()
	for _, deviceType := range devicetype.AllDeviceTypes {
		times[deviceType] = t
	}
	return times
}

// GetTimeForDevicetype -- Get the time for a given DeviceTypeId, or nil if not defined yet
//
//goland:noinspection GoMixedReceiverTypes
func (dttm DeviceTypeToTimeMap) GetTimeForDevicetype(dtype devicetype.DeviceType) *time.Time {
	timestamp, found := dttm[dtype]
	if found {
		return &timestamp
	} else {
		return nil
	}
}

// Scan -- Convert incoming bytes into a DeviceTypeToTimeMap structure
//
//goland:noinspection GoMixedReceiverTypes
func (dttm *DeviceTypeToTimeMap) Scan(value interface{}) error {
	bytes, ok := value.([]byte)
	if !ok {
		return errors.New(fmt.Sprint("Failed to unmarshal JSONB value:", value))
	}
	return json.Unmarshal(bytes, dttm)
}

// Value -- return json value of DeviceTypeToTimeMap, implement driver.Value interface
//
//goland:noinspection GoMixedReceiverTypes
func (dttm DeviceTypeToTimeMap) Value() (driver.Value, error) {
	if len(dttm) == 0 {
		return nil, nil
	}
	rawBytes, err := json.Marshal(dttm)
	if err != nil {
		return nil, err
	}
	return json.RawMessage(rawBytes).MarshalJSON()
}

// NewDeviceTypeChangeSet -- generate a new device type set based on a supplied map (The inverse of toMap)
//   - Converts a 'map[string]millis' structure into a 'map[DeviceTypeId]timestamp' structure
//
// TODO: May belong in the migrator package -- not used by hint generator
func NewDeviceTypeChangeSet(changes map[string]int64, source zap.Field) DeviceTypeToTimeMap {
	deviceTypeTimestamps := make(DeviceTypeToTimeMap)
	for deviceTypeName, timeLong := range changes {
		deviceType := devicetype.FromString(deviceTypeName)
		convertedTime := time.UnixMilli(timeLong)
		deviceTypeTimestamps[deviceType] = convertedTime
		if !deviceType.IsKnown() {
			logger.BaseLogger().Warn(fmt.Sprintf("Unknown DeviceType '%s' seen for Org DeviceTypeToTimeMap", deviceType),
				source)
		}
	}
	return deviceTypeTimestamps
}

// NewDeviceTypeChangeSetFromInterface -- generate a new device type set based on a supplied map
//   - Converts a 'map[string]int64' or 'map[string]float64' structure into a 'map[DeviceTypeId]timestamp' structure
func NewDeviceTypeChangeSetFromInterface(changes map[string]any, source zap.Field) DeviceTypeToTimeMap {
	int64Changes := make(map[string]int64, len(changes))
	for key, value := range changes {
		int64Changes[key] = int64(value.(float64))
	}

	return NewDeviceTypeChangeSet(int64Changes, source)
}
