// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package utils

import (
	"strings"
	"sync"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/logger"

	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
	"go.uber.org/zap/zapcore"
)

type SyncedCondBool struct {
	status    bool
	condition *sync.Cond
	lock      *sync.Mutex
}

func (scb *SyncedCondBool) Get() bool {
	scb.Lock()
	defer scb.lock.Unlock()
	return scb.status
}

func (scb *SyncedCondBool) Set(newValue bool) {
	scb.Lock()
	defer scb.lock.Unlock()
	scb.status = newValue
	scb.condition.Broadcast()
}

func (scb *SyncedCondBool) Lock() {
	scb.lock.Lock()
}

func (scb *SyncedCondBool) Wait() {
	scb.condition.Wait()
}

func NewSyncBool(initialValue bool) *SyncedCondBool {
	lock := sync.Mutex{}
	cond := sync.NewCond(&lock)
	return &SyncedCondBool{status: initialValue, condition: cond, lock: &lock}
}

type OutboundServiceConfig struct {
	Application string
	Host        string
	Component   string
	Category    string
}

// FIXME: SetOutboundServiceTags() unused, commented out for review
// SetOutboundServiceTags takes a span and an OutboundServiceConfig pointer and adds the appropriate outbound service tags to the span
// func SetOutboundServiceTags(span opentracing.Span, cfg *OutboundServiceConfig) {
//	span.SetTag("_outboundExternalService", cfg.Application)
//	span.SetTag("_externalApplication", "psc")
//	span.SetTag("_externalHost", cfg.Host)
//	span.SetTag("_externalComponent", cfg.Component)
//	span.SetTag("_externalCategory", cfg.Category)
//}

// FIXME: StartTracingSpan() unused, commented out for review
// StartTracingSpan -- FIXME: Describe this please
//
//goland:noinspection GoUnusedExportedFunction
// func StartTracingSpan(c context.Context, operationName, application, host, component, category string) (sp opentracing.Span, ctx context.Context) {
//	sp, ctx = opentracing.StartSpanFromContext(c, operationName)
//	if sp != nil {
//		SetOutboundServiceTags(sp, &OutboundServiceConfig{
//			Application: application,
//			Host:        host,
//			Component:   component,
//			Category:    category,
//		})
//	}
//	return
//}

// ValidateOrgKey -- Check if orgkey is valid, return key string and underlying env if valid
func ValidateOrgKey(orgKeyRaw string) (orgkey, env string, err error) {
	orgKey, err := orgkeylib.Decode(orgKeyRaw)
	if err == nil {
		return orgKeyRaw, orgKey.Env(), nil
	}
	return
}

// Sanitize -- Check to remove leading/trailing whitespace from a string; place warning in log if required
func Sanitize(what, source string) (safe string) {
	sanitized := strings.TrimSpace(source)
	if sanitized != source {
		logger.BaseLogger().Warn("Supplied '" + what + "' had leading/trailing whitespace that will be trimmed")
	}
	return sanitized
}

// GetTimePointer -- convenience method to get supplied time in millies as pointer
func GetTimePointer(milliSinceEpoch int64) *time.Time {
	value := time.UnixMilli(milliSinceEpoch)
	return &value
}

// GetTimePointerFromPointer -- convenience method to get supplied time in millies as pointer
func GetTimePointerFromPointer(milliSinceEpoch *int64) *time.Time {
	if milliSinceEpoch != nil {
		value := time.UnixMilli(*milliSinceEpoch)
		return &value
	}
	return nil
}

// CompareTimePointers -- Compare the values of two pointers to timestamps, return true if same time or both nil
func CompareTimePointers(one, two *time.Time) (equal bool) {
	if one != nil && two != nil {
		return one.Equal(*two)
	}
	if one == nil && two == nil {
		return true
	}
	return false
}

// AddTimePointerToEncoder -- add timestamp with label to zap logger
func AddTimePointerToEncoder(encoder zapcore.ObjectEncoder, label string, timePointer *time.Time) {
	if timePointer != nil {
		encoder.AddTime(label, *timePointer)
	}
}
