// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package utils

import (
	"testing"
	"time"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Utilities unit tests", func() {
	Context("ValidateOrgKey() testing", func() {
		It("should return info with proper orgkey", func() {
			source := "53VSWLAG8SYPF8JUQRTKU52P"
			check, env, err := ValidateOrgKey(source)

			Expect(err).ShouldNot(HaveOccurred())
			Expect(check).ToNot(BeNil())
			Expect(check).To(BeEquivalentTo(source))
			Expect(env).To(BeEquivalentTo("enki01"))
		})

		It("should return nil with invalid orgkey", func() {
			source := "BOGUS"
			check, env, err := ValidateOrgKey(source)

			Expect(err).Should(HaveOccurred())
			Expect(check).To(BeEquivalentTo(""))
			Expect(env).To(BeEquivalentTo(""))
		})
	})

	Context("Sanitize() testing", func() {
		It("should return trimmed string and raise warning if trim was done", func() {
			check := Sanitize("test", "  TRIM-ME  ")
			Expect(check).To(BeEquivalentTo("TRIM-ME"))
		})
	})

	Context("CompareTimePointers() testing", func() {
		It("should return true if both are nil", func() {
			check := CompareTimePointers(nil, nil)
			Expect(check).To(BeEquivalentTo(true))
		})
		It("should return true if both pointer values have the same time", func() {
			now1 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			now2 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			check := CompareTimePointers(&now1, &now2)
			Expect(check).To(BeEquivalentTo(true))
		})
		It("should return false if both pointer values have different times", func() {
			now1 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			now2 := time.Date(2023, 5, 19, 10, 15, 30, 1, time.UTC) // 1 ns difference
			check := CompareTimePointers(&now1, &now2)
			Expect(check).To(BeEquivalentTo(false))
		})
		It("should return false if one pointer is nil", func() {
			now1 := time.Date(2023, 5, 19, 10, 15, 30, 0, time.UTC)
			check := CompareTimePointers(&now1, nil)
			Expect(check).To(BeEquivalentTo(false))
		})
	})
	Context("HeaderStr() testing", func() {
		It("should return trailing ===== no matter the length", func() {
			check := HeaderStr("xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx")
			Expect(check).To(BeEquivalentTo("===== xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx ====="))
		})
	})
})

func TestUtils(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "All Utils Test Suite")
}
